// Copyright 2017 Google Inc. All Rights Reserved.
//
// Use of this source code is governed by an MIT-style
// license that can be found in the LICENSE file or at
// https://opensource.org/licenses/MIT.

#include "alpha.h"

#include <cstddef>
#include <cstdint>
#include <cstdio>

#include <memory>
#include <vector>

#include "ans_decode.h"
#include "bit_reader.h"
#include "dc_predictor.h"
#include "entropy_coder.h"
#include "lossless16.h"
#include "lossless8.h"
#include "status.h"

namespace pik {

namespace {

// TODO(veluca): check if those upper bounds can be improved.
const constexpr int kRleSymStart[2] = {10, 18};

const constexpr int kMaxRleBits = 31;
const constexpr int kMaxRleLength =
    DecodeVarLenUint(kMaxRleBits, (1u << kMaxRleBits) - 1);

}  // namespace

/* EncodeAlpha() removed */

Status DecodeAlpha(const DecompressParams& params, const Alpha& alpha,
                   ImageU* plane, const Rect& rect) {
  PIK_CHECK(plane->xsize() != 0);
  if (alpha.bytes_per_alpha != 1 && alpha.bytes_per_alpha != 2) {
    return PIK_FAILURE("Invalid bytes_per_alpha");
  }

  BitReader bit_reader(alpha.encoded.data(), alpha.encoded.size());
  std::vector<uint8_t> context_map;
  ANSCode code;
  PIK_RETURN_IF_ERROR(
      DecodeHistograms(&bit_reader, 1, 54, &code, &context_map));
  ANSSymbolReader decoder(&code);

  const size_t xsize = rect.xsize();
  const size_t ysize = rect.ysize();
  const size_t rle_sym_start = kRleSymStart[alpha.bytes_per_alpha - 1];

  ImageS residuals(xsize, ysize);
  const int histo_idx = context_map[0];
  size_t skip = 0;
  for (size_t y = 0; y < ysize; y++) {
    int16_t* PIK_RESTRICT row = residuals.Row(y);
    for (size_t x = 0; x < xsize; x++) {
      if (skip) {
        row[x] = 0;
        skip--;
        continue;
      }
      bit_reader.FillBitBuffer();
      int s = decoder.ReadSymbol(histo_idx, &bit_reader);
      if (s > 0) {
        if (s >= rle_sym_start) {
          s -= rle_sym_start;
          if (s > kMaxRleBits) {
            return PIK_FAILURE("Invalid rle nbits");
          }
          int bits = bit_reader.ReadBits(s);
          s = DecodeVarLenUint(s, bits);
          skip = s;
          row[x] = 0;
          continue;
        }
        int bits = bit_reader.ReadBits(s);
        s = DecodeVarLenInt(s, bits);
      }
      row[x] = s;
    }
  }
  if (skip != 0) {
    return PIK_FAILURE("Invalid alpha");
  }
  PIK_RETURN_IF_ERROR(bit_reader.JumpToByteBoundary());
  ImageS alpha_img(xsize, ysize);
  ExpandY(Rect(alpha_img), residuals, &alpha_img);
  if (alpha.bytes_per_alpha == 2) {
    for (size_t y = 0; y < ysize; y++) {
      const int16_t* PIK_RESTRICT in = alpha_img.ConstRow(y);
      uint16_t* PIK_RESTRICT row = rect.Row(plane, y);
      for (size_t x = 0; x < rect.xsize(); x++) {
        row[x] = in[x] + (1 << 15);
      }
    }
  } else {
    for (size_t y = 0; y < ysize; y++) {
      const int16_t* PIK_RESTRICT in = alpha_img.ConstRow(y);
      uint16_t* PIK_RESTRICT row = rect.Row(plane, y);
      for (size_t x = 0; x < rect.xsize(); x++) {
        row[x] = in[x];
      }
    }
  }

  if (!decoder.CheckANSFinalState()) {
    return PIK_FAILURE("ANS checksum failure.");
  }
  return true;
}

}  // namespace pik
