// Copyright 2017 Google Inc. All Rights Reserved.
//
// Use of this source code is governed by an MIT-style
// license that can be found in the LICENSE file or at
// https://opensource.org/licenses/MIT.

#include "block_dictionary.h"
#include <sys/types.h>
#include <cstdint>
#include <limits>
#include <vector>
#include "codec.h"
#include "entropy_coder.h"
#include "image.h"
#include "opsin_inverse.h"
#include "status.h"

namespace pik {
constexpr float kBlockScale = 3.5;
constexpr float kBlockInvScale = 1.0f / kBlockScale;
constexpr int kMaxBlocks = 1 << 24;

// TODO(veluca): choose some reasonable set of blocks as a static dictionary.
constexpr int kNumStaticBlocks = 1;
QuantizedBlock kStaticBlocks[kNumStaticBlocks] = {
    {1, 3, {{0, 0, 0}, {-11, -8, -11}, {-10, -7, -10}}},
};

enum Contexts {
  kNumBlockContext = 0,
  kBlockSizeContext = 1,
  kPixelsContextStart = 2,
  kPixelsContextY = 3,
  kPixelsContextB = 4,
  kBlockOffsetContext = 5,
  kBlockWidthContext = 6,
  kBlockIdCountContext = 7,
  kBlockIdSkipContext = 8,
  kNumBlockDictionaryContexts,
};

// We can represent numbers up to 2**kMaxBlockDictionarySymbol. As the biggest
// numbers have an image-size range, 2**32 should be more than enough here.
constexpr int kMaxBlockDictionarySymbol = 32;

/* ScaleForQuantization() removed */
/* Quantize() removed */

// BlockInfo contains the data describing a block before quantization
// and also contains its position.
struct BlockInfo {
  Rect rect;
  const Image3F* image;
  BlockInfo(const Image3F* image, Rect rect) : rect(rect), image(image) {}

  /* QuantizedBlock() removed */

  bool IsSimilar(const BlockInfo& other) const {
    if (other.rect.xsize() != rect.xsize()) return false;
    if (other.rect.ysize() != rect.ysize()) return false;
    float sum = 0;
    for (size_t c = 0; c < 3; c++) {
      for (size_t iy = 0; iy < rect.ysize(); iy++) {
        const float* PIK_RESTRICT row = image->ConstPlaneRow(c, rect.y0() + iy);
        const float* PIK_RESTRICT other_row =
            other.image->ConstPlaneRow(c, other.rect.y0() + iy);
        for (size_t ix = 0; ix < rect.xsize(); ix++) {
          float diff = row[rect.x0() + ix] - other_row[other.rect.x0() + ix];
          sum += diff * diff;
        }
      }
    }
    if (sum > kDotDistThreshold) return false;

    return true;
  }
};

BlockDictionary::BlockDictionary(const std::vector<QuantizedBlock>& dictionary,
                                 const std::vector<BlockPosition>& positions)
    : dictionary_(dictionary), positions_(positions) {
  std::sort(positions_.begin(), positions_.end(),
            [](const BlockPosition& a, const BlockPosition& b) {
              return std::make_tuple(a.transform, a.id, a.x, a.y, a.dx, a.dy,
                                     a.width) <
                     std::make_tuple(b.transform, b.id, b.x, b.y, b.dx, b.dy,
                                     b.width);
            });
}

/* BlockDictionary::Encode() removed */

Status BlockDictionary::Decode(BitReader* br, size_t xsize, size_t ysize) {
  std::vector<uint8_t> context_map;
  ANSCode code;
  PIK_RETURN_IF_ERROR(DecodeHistograms(br, kNumBlockDictionaryContexts,
                                       kMaxBlockDictionarySymbol, &code,
                                       &context_map));
  ANSSymbolReader decoder(&code);

  auto read_num = [&](int context) {
    br->FillBitBuffer();
    int s = decoder.ReadSymbol(context_map[context], br);
    int bits = br->ReadBits(s);
    return DecodeVarLenUint(s, bits);
  };

  size_t dict_size = read_num(kNumBlockContext);
  if (dict_size > kMaxBlocks) {
    return PIK_FAILURE("Too many blocks in dictionary");
  }

  dictionary_.resize(dict_size);
  for (size_t i = 0; i < dictionary_.size(); i++) {
    QuantizedBlock& info = dictionary_[i];
    info.xsize = read_num(kBlockSizeContext);
    info.ysize = read_num(kBlockSizeContext);
    if (i != 0) {
      info.xsize = UnpackSigned(info.xsize) + dictionary_[i - 1].xsize;
      info.ysize = UnpackSigned(info.ysize) + dictionary_[i - 1].ysize;
    }
    if (info.xsize > kMaxBlockSize)
      return PIK_FAILURE("Block xsize is too big: %lu", info.xsize);
    if (info.ysize > kMaxBlockSize)
      return PIK_FAILURE("Block ysize is too big: %lu", info.ysize);
    for (size_t c = 0; c < 3; c++) {
      int ctx = kPixelsContextStart + c;
      for (size_t iy = 0; iy < info.ysize; iy++) {
        for (size_t ix = 0; ix < info.xsize; ix++) {
          int8_t pred = 0;
          if (ix != 0) {
            pred += info.pixels[c][iy * info.xsize + ix - 1];
          }
          if (iy != 0) {
            pred += info.pixels[c][(iy - 1) * info.xsize + ix];
          }
          if (ix != 0 && iy != 0) pred >>= 1;
          info.pixels[c][iy * info.xsize + ix] =
              UnpackSigned(read_num(ctx)) + pred;
        }
      }
    }
  }

  for (size_t transform = 0; transform < 2; transform++) {
    size_t to_skip = 0;
    for (size_t id = 0; id < dictionary_.size() + kNumStaticBlocks; id++) {
      if (to_skip > 0) {
        to_skip--;
        continue;
      }
      size_t id_count = read_num(kBlockIdCountContext);
      if (id_count > kMaxBlocks) {
        return PIK_FAILURE("Too many blocks in dictionary");
      }
      if (id_count == 0) {
        to_skip = read_num(kBlockIdSkipContext);
        continue;
      }
      positions_.resize(positions_.size() + id_count);
      size_t id_start = positions_.size() - id_count;
      for (size_t i = id_start; i < positions_.size(); i++) {
        BlockPosition& pos = positions_[i];
        pos.x = read_num(kBlockOffsetContext);
        pos.y = read_num(kBlockOffsetContext);
        pos.id = id;
        const QuantizedBlock& info =
            pos.id < dictionary_.size()
                ? dictionary_[pos.id]
                : kStaticBlocks[pos.id - dictionary_.size()];
        if (i != id_start) {
          pos.x = UnpackSigned(pos.x) + positions_[i - 1].x;
          pos.y = UnpackSigned(pos.y) + positions_[i - 1].y;
        }
        if (pos.x + info.xsize > xsize) {
          return PIK_FAILURE("Invalid block x (id %lu): at %lu + %lu > %lu",
                             pos.id, pos.x, info.xsize, xsize);
        }
        if (pos.y + info.ysize > ysize) {
          return PIK_FAILURE("Invalid block y: at %lu + %lu > %lu", pos.y,
                             info.ysize, ysize);
        }
        pos.transform = transform;
        if (transform) {
          pos.dx = UnpackSigned(read_num(kBlockOffsetContext));
          pos.dy = UnpackSigned(read_num(kBlockOffsetContext));
          pos.width = UnpackSigned(read_num(kBlockWidthContext));
        }
      }
    }
    if (to_skip > 0) {
      return PIK_FAILURE("Invalid number of skipped block ids!");
    }
  }

  if (!decoder.CheckANSFinalState()) {
    return PIK_FAILURE("ANS checksum failure.");
  }
  PIK_RETURN_IF_ERROR(br->JumpToByteBoundary());
  return true;
}

/* DumpImage() removed */

template <bool add>
void BlockDictionary::Apply(Image3F* opsin, size_t downsampling) const {
  if (downsampling != 1) {
    // TODO(veluca): downsampling not implemented yet.
    PIK_CHECK(positions_.empty());
  }
  // Blocks copied as-is.
  for (const BlockPosition& pos : positions_) {
    if (pos.transform) continue;
    size_t by = pos.y;
    size_t bx = pos.x;
    const QuantizedBlock& info =
        pos.id < dictionary_.size()
            ? dictionary_[pos.id]
            : kStaticBlocks[pos.id - dictionary_.size()];
    for (size_t c = 0; c < 3; c++) {
      for (size_t iy = 0; iy < info.ysize; iy++) {
        float* row = opsin->PlaneRow(c, by + iy) + bx;
        for (size_t ix = 0; ix < info.xsize; ix++) {
          float val = kBlockInvScale * kXybRadius[c] *
                      info.pixels[c][iy * info.xsize + ix];
          if (add) {
            row[ix] += val;
          } else {
            row[ix] -= val;
          }
        }
      }
    }
  }
  // Scaled/rotated blocks.
  float* PIK_RESTRICT rows[3] = {opsin->PlaneRow(0, 0), opsin->PlaneRow(1, 0),
                                 opsin->PlaneRow(2, 0)};
  size_t stride = opsin->PixelsPerRow();
  for (const BlockPosition& pos : positions_) {
    if (!pos.transform) continue;
    float block[3][(2 * kMaxBlockSize + 5) * (2 * kMaxBlockSize + 5)] = {};
    const QuantizedBlock& info =
        pos.id < dictionary_.size()
            ? dictionary_[pos.id]
            : kStaticBlocks[pos.id - dictionary_.size()];
    size_t xs = info.xsize;
    size_t ys = info.ysize;
    for (size_t c = 0; c < 3; c++) {
      for (int iy = 0; iy < int(2 * ys) + 1; iy++) {
        for (int ix = 0; ix < int(2 * xs) + 1; ix++) {
          float val = 0.0f;
          for (int dy = -1; dy < 1; dy++) {
            for (int dx = -1; dx < 1; dx++) {
              int sy = (iy + dy) / 2;
              if (sy < 0) sy = 0;
              if (sy >= ys) sy = ys - 1;
              int sx = (ix + dx) / 2;
              if (sx < 0) sx = 0;
              if (sx >= xs) sx = xs - 1;
              val += info.pixels[c][sy * info.xsize + sx];
            }
          }
          block[c][(iy + 2) * (2 * xs + 5) + (ix + 2)] = val * 0.25f;
        }
      }
    }
    float by00 = pos.y;
    float bx00 = pos.x;
    float bx01 = bx00 + pos.dx;
    float by01 = by00 + pos.dy;
    float xnorm = std::sqrt(pos.dx * pos.dx + pos.dy * pos.dy);
    float ynorm = pos.width * 0.5f;
    float invxnorm = 1.0f / xnorm;
    float invynorm = 1.0f / ynorm;
    float deltax = pos.dy * invxnorm * ynorm;
    float deltay = -pos.dx * invxnorm * ynorm;
    float bx11 = bx01 + deltax;
    float by11 = by01 + deltay;
    float bx10 = bx00 + deltax;
    float by10 = by00 + deltay;
    float inv_determinant = 1.0f / (pos.dx * deltay - pos.dy * deltax);
    float inverse_transform[4] = {
        xs * deltay * inv_determinant, -float(xs) * deltax * inv_determinant,
        -float(ys) * pos.dy * inv_determinant, ys * pos.dx * inv_determinant};
    int64_t min_x = std::min(std::min(bx00, bx01), std::min(bx10, bx11));
    int64_t min_y = std::min(std::min(by00, by01), std::min(by10, by11));
    int64_t max_x = std::max(std::max(bx00, bx01), std::max(bx10, bx11)) + 1;
    int64_t max_y = std::max(std::max(by00, by01), std::max(by10, by11)) + 1;
    if (min_x < 0) min_x = 0;
    if (min_y < 0) min_y = 0;
    if (max_x > opsin->xsize()) max_x = opsin->xsize();
    if (max_y > opsin->ysize()) max_y = opsin->ysize();
    constexpr float kAntialiasingMargin = 0.3f;
    float margin_x = kAntialiasingMargin * invxnorm;
    float margin_y = kAntialiasingMargin * invynorm;
    for (size_t iy = min_y; iy < max_y; iy++) {
      for (size_t ix = min_x; ix < max_x; ix++) {
        float x = ix - bx00 + 0.5f;
        float y = iy - by00 + 0.5f;
        float ox = inverse_transform[0] * x + inverse_transform[1] * y;
        float oy = inverse_transform[2] * x + inverse_transform[3] * y;
        if (ox >= -margin_x && ox < xs + margin_x && oy >= -margin_y &&
            oy < ys + margin_y) {
          ox = 2 * ox + 2;
          oy = 2 * oy + 2;
          int floorx = ox;
          int ceilx = floorx + 1;
          float fracx = ox - floorx;
          int floory = oy;
          int ceily = floory + 1;
          float fracy = oy - floory;
          for (size_t c = 0; c < 3; c++) {
            float val =
                (fracx * fracy * block[c][ceily * (2 * xs + 5) + ceilx] +
                 fracx * (1.f - fracy) *
                     block[c][floory * (2 * xs + 5) + ceilx] +
                 (1.f - fracx) * fracy *
                     block[c][ceily * (2 * xs + 5) + floorx] +
                 (1.f - fracx) * (1.f - fracy) *
                     block[c][floory * (2 * xs + 5) + floorx]) *
                kBlockInvScale * kXybRadius[c];
            if (add) {
              rows[c][iy * stride + ix] += val;
            } else {
              rows[c][iy * stride + ix] -= val;
            }
          }
        }
      }
    }
  }
}

void BlockDictionary::AddTo(Image3F* opsin, size_t downsampling) const {
  Apply</*add=*/true>(opsin, downsampling);
}

void BlockDictionary::SubtractFrom(Image3F* opsin) const {
  Apply</*add=*/false>(opsin, /*downsampling=*/1);
}

namespace {

/* Distance() removed */

constexpr size_t kNumExploreSteps = 1;
constexpr size_t kSmallBlockThreshold = 13;
constexpr float kDistThreshold = 0.6f;
// Returns north, south, east and west neighbors, if present.
size_t Neighbors(size_t x, size_t y, size_t x_max, size_t y_max,
                 std::array<size_t, 2>* neighbors) {
  size_t i = 0;
  if (x != 0) neighbors[i++] = {x - 1, y};
  if (y != 0) neighbors[i++] = {x, y - 1};
  if (x != x_max - 1) neighbors[i++] = {x + 1, y};
  if (y != y_max - 1) neighbors[i++] = {x, y + 1};
  return i;
}

struct Site {
  size_t x;
  size_t y;
  size_t steps;
  constexpr Site(size_t x, size_t y, size_t steps) : x(x), y(y), steps(steps) {}
};

// Finds a bounding box for a loosly connected component. If steps==1, only
// active neighboring pixels are added to the component. If `steps` is larger,
// it is allowed to take a few `steps` through non-active neighboring pixels.
Rect ConnectedCompenentBounds(ImageI* PIK_RESTRICT active, size_t x, size_t y) {
  Rect box = Rect(x, y, 0, 0);
  std::vector<Site> places_to_visit;
  size_t steps = kNumExploreSteps;
  Site site(x, y, steps);
  places_to_visit.emplace_back(site);

  while (!places_to_visit.empty()) {
    Site site = places_to_visit.back();
    x = site.x;
    y = site.y;
    steps = site.steps;
    places_to_visit.pop_back();

    if (steps == 0) continue;
    uint8_t cell_type = active->ConstRow(y)[x];
    if (cell_type == 2) continue;

    if (cell_type == 1) {
      size_t xmin = std::min(x, box.x0());
      size_t ymin = std::min(y, box.y0());
      size_t xmax = std::max(x + 1, box.x0() + box.xsize());
      size_t ymax = std::max(y + 1, box.y0() + box.ysize());
      if (((xmax - xmin) < kMaxBlockSize) && ((ymax - ymin) < kMaxBlockSize)) {
        box = Rect(xmin, ymin, xmax - xmin, ymax - ymin);
      } else {
        continue;
      }
    }
    std::array<size_t, 2> neighbors[4];
    for (int i = 0;
         i < Neighbors(x, y, active->xsize(), active->ysize(), neighbors);
         i++) {
      size_t new_x = neighbors[i][0];
      size_t new_y = neighbors[i][1];
      active->Row(y)[x] = 2;
      site = {new_x, new_y, cell_type ? kNumExploreSteps : steps - 1};
      places_to_visit.emplace_back(site);
    }
  }
  return box;
}
};  // namespace

static const bool kUseBlockDictionary = false;
static const bool kUseHardcodedStretchedBlocks = false;
static const bool KUseDotDetection = true;

/* FindBestBlockDictionary() removed */
}  // namespace pik
