// Copyright 2019 Google Inc. All Rights Reserved.
//
// Use of this source code is governed by an MIT-style
// license that can be found in the LICENSE file or at
// https://opensource.org/licenses/MIT.

#ifndef PIK_IMAGE_OPS_H_
#define PIK_IMAGE_OPS_H_

// Operations on images.

#include "image.h"

namespace pik {

/* Subtract() removed */

/* SubtractFrom() removed */

/* AddTo() removed */

/* LinComb() removed */

/* ScaleImage() removed */

/* Product() removed */

/* DotProduct() removed */

template <typename T>
void FillImage(const T value, Image<T>* image) {
  for (size_t y = 0; y < image->ysize(); ++y) {
    T* const PIK_RESTRICT row = image->Row(y);
    for (size_t x = 0; x < image->xsize(); ++x) {
      row[x] = value;
    }
  }
}

template <typename T>
void ZeroFillImage(Image<T>* image) {
  for (size_t y = 0; y < image->ysize(); ++y) {
    T* const PIK_RESTRICT row = image->Row(y);
    memset(row, 0, image->xsize() * sizeof(T));
  }
}

/* class GeneratorRandom removed */

/* class GeneratorRandom<float, Random> removed */

/* class GeneratorRandom<double, Random> removed */

/* GenerateImage() removed */

/* RandomFillImage() removed */

/* RandomFillImage() removed */

// Mirrors out of bounds coordinates and returns valid coordinates unchanged.
// We assume the radius (distance outside the image) is small compared to the
// image size, otherwise this might not terminate.
// The mirror is outside the last column (border pixel is also replicated).
static inline int64_t Mirror(int64_t x, const int64_t xsize) {
  // TODO(janwas): replace with branchless version
  while (x < 0 || x >= xsize) {
    if (x < 0) {
      x = -x - 1;
    } else {
      x = 2 * xsize - 1 - x;
    }
  }
  return x;
}

// Wrap modes for ensuring X/Y coordinates are in the valid range [0, size):

// Mirrors (repeating the edge pixel once). Useful for convolutions.
struct WrapMirror {
  PIK_INLINE int64_t operator()(const int64_t coord, const int64_t size) const {
    return Mirror(coord, size);
  }
};

// Repeats the edge pixel.
struct WrapClamp {
  PIK_INLINE int64_t operator()(const int64_t coord, const int64_t size) const {
    return std::min(std::max<int64_t>(0, coord), size - 1);
  }
};

// Returns the same coordinate: required for TFNode with Border(), or useful
// when we know "coord" is already valid (e.g. interior of an image).
struct WrapUnchanged {
  PIK_INLINE int64_t operator()(const int64_t coord, const int64_t size) const {
    return coord;
  }
};

// Similar to Wrap* but for row pointers (reduces Row() multiplications).

class WrapRowMirror {
 public:
  template <class ImageOrView>
  WrapRowMirror(const ImageOrView& image, const size_t ysize)
      : first_row_(image.ConstRow(0)), last_row_(image.ConstRow(ysize - 1)) {}

  const float* const PIK_RESTRICT
  operator()(const float* const PIK_RESTRICT row, const int64_t stride) const {
    if (row < first_row_) {
      const int64_t num_before = first_row_ - row;
      // Mirrored; one row before => row 0, two before = row 1, ...
      return first_row_ + num_before - stride;
    }
    if (row > last_row_) {
      const int64_t num_after = row - last_row_;
      // Mirrored; one row after => last row, two after = last - 1, ...
      return last_row_ - num_after + stride;
    }
    return row;
  }

 private:
  const float* const PIK_RESTRICT first_row_;
  const float* const PIK_RESTRICT last_row_;
};

struct WrapRowUnchanged {
  PIK_INLINE const float* const PIK_RESTRICT
  operator()(const float* const PIK_RESTRICT row, const int64_t stride) const {
    return row;
  }
};

/* SetBorder() removed */

/* ImageMinMax() removed */

/* ImageAverage() removed */

/* ImageConvert() removed */

/* StaticCastImage() removed */


/* ImageFromPacked() removed */

/* Image3MinMax() removed */

/* Image3MinMax() removed */

/* Image3Max() removed */

/* ImageSum() removed */

/* ImageSum() removed */

/* Image3Sum() removed */

/* Image3Sum() removed */

/* PackedFromImage() removed */

/* PackedFromImage() removed */

/* ImageMedian() removed */

/* ImageMedian() removed */

/* Image3Median() removed */

/* Image3Median() removed */

/* Image3Convert() removed */

/* StaticCastImage3() removed */

template <typename Tin, typename Tout>
void Subtract(const Image3<Tin>& image1, const Image3<Tin>& image2,
              Image3<Tout>* out) {
  const size_t xsize = image1.xsize();
  const size_t ysize = image1.ysize();
  PIK_CHECK(xsize == image2.xsize());
  PIK_CHECK(ysize == image2.ysize());

  for (int c = 0; c < 3; ++c) {
    for (size_t y = 0; y < ysize; ++y) {
      const Tin* const PIK_RESTRICT row1 = image1.ConstPlaneRow(c, y);
      const Tin* const PIK_RESTRICT row2 = image2.ConstPlaneRow(c, y);
      Tout* const PIK_RESTRICT row_out = out->PlaneRow(c, y);
      for (size_t x = 0; x < xsize; ++x) {
        row_out[x] = row1[x] - row2[x];
      }
    }
  }
}

/* SubtractFrom() removed */

/* AddTo() removed */

/* ScaleImage() removed */

/* FillImage() removed */

template <typename T>
void ZeroFillImage(Image3<T>* image) {
  for (int c = 0; c < 3; ++c) {
    for (size_t y = 0; y < image->ysize(); ++y) {
      T* PIK_RESTRICT row = image->PlaneRow(c, y);
      memset(row, 0, image->xsize() * sizeof(T));
    }
  }
}

template <typename T>
// TODO(firsching): add rect parameter to ZeroFillImage for consistency.
void ZeroFillImage(Image<T>* image, Rect rect) {
  for (size_t y = 0; y < rect.ysize(); ++y) {
    T* PIK_RESTRICT row = rect.Row(image, y);
    memset(row, 0, rect.xsize() * sizeof(T));
  }
}

/* GenerateImage() removed */

/* RandomFillImage() removed */

/* RandomFillImage() removed */

/* InterleavedFromImage3() removed */

/* Image3FromInterleaved() removed */

// First, image is padded horizontally, with the rightmost value.
// Next, image is padded vertically, by repeating the last line.
Image3F PadImageToMultiple(const Image3F& in, const size_t N);

}  // namespace pik

#endif  // PIK_IMAGE_OPS_H_
