// Copyright 2018 Google Inc. All Rights Reserved.
//
// Use of this source code is governed by an MIT-style
// license that can be found in the LICENSE file or at
// https://opensource.org/licenses/MIT.

#ifndef PIK_MULTIPASS_HANDLER_H_
#define PIK_MULTIPASS_HANDLER_H_

#include "ac_strategy.h"
#include "block_dictionary.h"
#include "codec.h"
#include "color_correlation.h"
#include "color_encoding.h"
#include "common.h"
#include "compressed_image_fwd.h"
#include "data_parallel.h"
#include "headers.h"
#include "image.h"
#include "pik_params.h"
#include "quant_weights.h"
#include "quantizer.h"
#include "status.h"

// Defines how multi-pass images should be encoded and decoded.

namespace pik {

class MultipassManager;

// MultipassHandler is a child object of MultipassManager. It is bound to
// specific group (see GetGroupHandler) and is used to perform operations over
// that group region.
class MultipassHandler {
 public:
  virtual ~MultipassHandler() = default;

  virtual const Rect& GroupRect() = 0;
  virtual const Rect& PaddedGroupRect() = 0;
  Rect BlockGroupRect() {
    const Rect& r = PaddedGroupRect();
    return Rect(r.x0() / kBlockDim, r.y0() / kBlockDim, r.xsize() / kBlockDim,
                r.ysize() / kBlockDim);
  }

  // Progressive mode.
  virtual std::vector<Image3S> SplitACCoefficients(
      Image3S&& ac, const AcStrategyImage& ac_strategy) {
    std::vector<Image3S> ret;
    ret.push_back(std::move(ac));
    return ret;
  }

  // Returns the MultipassManager this handler was created by.
  virtual MultipassManager* Manager() = 0;

 private:
  DequantMatrices default_matrices_{/*need_inv_matrices=*/false};
  Quantizer quantizer_{&default_matrices_, 0, 0};
};

// MultipassManager holds information about passes and manages
// MultipassHandlers. It is assumed that parallelization goes below the manager
// level (at group level), so all the methods of MultipassManager should be
// invoked from a single thread.
class MultipassManager {
 public:
  virtual ~MultipassManager() = default;

  // Modifies img, applying a transformation that reduces its entropy given a
  // reference image, typically the output of a previous pass.
  virtual void DecorrelateOpsin(Image3F* img) = 0;

  // Inverse of DecorrelateOpsin.
  virtual void RestoreOpsin(Image3F* img) = 0;

  // Called at the start of each pass.
  virtual void StartPass(const FrameHeader& header) = 0;

  // Called by the decoder when a pass is done.
  virtual void SetDecodedPass(const Image3F& opsin) = 0;

  // This version is only called if we decoded a lossless pass.
  virtual void SetDecodedPass(CodecInOut* io) = 0;

  // Used *on the encoder only* to forcibly enable adaptive reconstruction in
  // GetQuantizer.
  virtual void UseAdaptiveReconstruction() {}

  // NOTE: not thread safe.
  // Preferably, `group_id` should be small non-negative number.
  // Same `group_rect` should be provided with corresponding `group_id`.
  virtual MultipassHandler* GetGroupHandler(size_t group_id,
                                            const Rect& group_rect) = 0;

  /* GetDequantMatrices() removed */

  /* GetBlockDictionary() removed */

  /* GetColorCorrelationMap() removed */

  /* GetAcStrategy() removed */

  /* getQuantizer() removed */

  virtual size_t GetNumPasses() { return 1; }
  virtual std::vector<std::pair<uint32_t, uint32_t>>
  GetDownsamplingToNumPasses() {
    return {};
  }

  // Save the ac strategy / quant field of this pass.
  virtual void SaveAcStrategy(const AcStrategyImage& af) {}
  virtual void SaveQuantField(const ImageI& qf) {}

  // Give a hint to the ac strategy / quant field encoder/decoder.
  virtual const AcStrategyImage* HintAcStrategy() { return nullptr; }
  virtual const ImageI* HintQuantField() { return nullptr; }

  // Previous pass in a specific colorspace.
  virtual Status GetPreviousPass(const ColorEncoding& color_encoding,
                                 ThreadPool* pool, Image3F* out) {
    return true;
  }
};

}  // namespace pik

#endif  // PIK_MULTIPASS_HANDLER_H_
