/*
 * Copyright (C) 2021 - 2025  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
extern "C"{
#define new new_
#include <abydos-plugin.h>
#undef new
#include <debug.h>
}

#include "codec.h"
#include "dpik.h"
#include "external_image.h"
#include "os_specific.h"
#include "pik_params.h"
#include "simd/targets.h"

#define MIN(a,b) ((a)<(b) ? (a) : (b))
#define MAX(a,b) ((a)>(b) ? (a) : (b))
#define CLAMP(x,lo,hi) MIN(hi, MAX(lo,x))

#define REPS 1

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};

static cairo_surface_t *
_surface_from_image_rgb(const pik::Image3F *img){
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, img->xsize(), img->ysize());
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface) / sizeof(uint32_t) - img->xsize();
    for(size_t y=0; y<img->ysize(); ++y){
        const float *row[3];
        row[0] = img->PlaneRow(0, y);
        row[1] = img->PlaneRow(1, y);
        row[2] = img->PlaneRow(2, y);
        for(size_t x=0; x<img->xsize(); ++x){
            int r = row[2][x];
            int g = row[1][x];
            int b = row[0][x];
            *dst++ = r << 0 | g << 8 | b << 16;
        }
        dst += dst_skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_image_rgba(const pik::Image3F *img, const pik::ImageU *alpha){
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, img->xsize(), img->ysize());
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface) / sizeof(uint32_t) - img->xsize();
    for(size_t y=0; y<img->ysize(); ++y){
        const float *row[3];
        const uint16_t *rowa;
        row[0] = img->PlaneRow(0, y);
        row[1] = img->PlaneRow(1, y);
        row[2] = img->PlaneRow(2, y);
        rowa = alpha->Row(y);
        for(size_t x=0; x<img->xsize(); ++x){
            int a = rowa[x];
            int r = a * CLAMP(row[2][x], 0, 255) / 255;
            int g = a * CLAMP(row[1][x], 0, 255) / 255;
            int b = a * CLAMP(row[0][x], 0, 255) / 255;
            *dst++ = r << 0 | g << 8 | b << 16 | a << 24;
        }
        dst += dst_skip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static cairo_surface_t *
_surface_from_io(pik::CodecInOut *io, pik::ThreadPool *pool){
    try{
        if(!io->TransformTo(io->dec_c_original, pool))
            return NULL;
    }catch(...){
        return NULL;
    }
    return io->HasAlpha() ? _surface_from_image_rgba(&io->color(), &io->alpha())
                          : _surface_from_image_rgb(&io->color());
}


static int
_create_from_data(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    const int bits = pik::TargetBitfield().Bits();
    if ((bits & SIMD_ENABLE) != SIMD_ENABLE) {
        DEBUG_p("CPU does not support all enabled targets");
        return -1;
    }

    pik::PaddedBytes bytes(len);
    char *dst = reinterpret_cast<char*>(&(bytes)[0]);
    memcpy(dst, data, len);

    pik::CodecContext codec_context;
    pik::ThreadPool pool(4);

    const std::vector<int> cpus = pik::AvailableCPUs();
    pool.RunOnEachThread([&cpus](const int task, const unsigned int thread){
        // 1.1-1.2x speedup (36 cores) from pinning.
        if (thread < cpus.size()) {
            if (!pik::PinThreadToCPU(cpus[thread])) {
                DEBUG_p("failed to pin thread %d", thread);
            }
        }
    });

    pik::CodecInOut io(&codec_context);
    size_t downsampling = 1;
    try{
        for(int i=REPS;i>0;--i)
            if(!pik::Decompress(bytes, pik::DecompressParams() /* args.params */, &pool, &io, &downsampling))
                return -1;
    }catch(...){
        return -1;
    }

    h->info->width = io.xsize();
    h->info->height = io.ysize();

    h->surface = _surface_from_io(&io, &pool);

    return h->surface ? 0 : -1;
}


extern "C" const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = [](){
            static const char *mime_types[] = {
                "image/x-pik",
                NULL
            };
            return mime_types;
        },
        .new_ = [](const char *mime_type, abydos_plugin_info_t *info){
            auto h = new abydos_plugin_handle_t;
            info->threadsafe = 1;
            h->info = info;
            h->surface = NULL;
            return h;
        },
        .free = [](abydos_plugin_handle_t *h){
            if(h->surface)
                cairo_surface_destroy(h->surface);
            delete h;
        },
        .create_from_data = _create_from_data,
        .get_image_surface = [](abydos_plugin_handle_t *h,
                                int page,
                                cairo_bool_t *layer_visibility,
                                int variant,
                                double time,
                                int frame)
        {
            return cairo_surface_reference(h->surface);
        }
    };
    return &p;
}
