// Copyright 2017 Google Inc. All Rights Reserved.
//
// Use of this source code is governed by an MIT-style
// license that can be found in the LICENSE file or at
// https://opensource.org/licenses/MIT.

#include "pik.h"

#include <string>
#include <vector>

#include "common.h"
#include "compressed_image.h"
#include "headers.h"
#include "image.h"
#include "multipass_handler.h"
#include "noise.h"
#include "os_specific.h"
#include "pik_frame.h"
#include "pik_params.h"
#include "quantizer.h"
#include "single_image_handler.h"

namespace pik {

namespace {
static const uint8_t kBrunsliMagic[] = {0x0A, 0x04, 'B', 0xd2, 0xd5, 'N', 0x12};

// TODO(user): use VerifySignature, when brunsli codebase is attached.
bool IsBrunsliFile(const PaddedBytes& compressed) {
  const size_t magic_size = sizeof(kBrunsliMagic);
  if (compressed.size() < magic_size) {
    return false;
  }
  if (memcmp(compressed.data(), kBrunsliMagic, magic_size) != 0) {
    return false;
  }
  return true;
}

Status BrunsliToPixels(const DecompressParams& dparams,
                       const PaddedBytes& compressed, CodecInOut* io,
                       PikInfo* aux_out, ThreadPool* pool) {
  return PIK_FAILURE("Brunsli decoding is not implemented yet.");
}

}  // namespace

/* PixelsToPik() removed */

Status PikToPixels(const DecompressParams& dparams,
                   const PaddedBytes& compressed, CodecInOut* io,
                   PikInfo* aux_out, ThreadPool* pool) {

  if (IsBrunsliFile(compressed)) {
    return BrunsliToPixels(dparams, compressed, io, aux_out, pool);
  }

  // To avoid the complexity of file I/O and buffering, we assume the bitstream
  // is loaded (or for large images/sequences: mapped into) memory.
  BitReader reader(compressed.data(), compressed.size());
  FileHeader container;
  PIK_RETURN_IF_ERROR(ReadFileHeader(&reader, &container));

  // Preview is discardable, i.e. content image does not rely on decoded preview
  // pixels; just skip it, if any.
  size_t preview_size_bits = container.preview.size_bits;
  if (preview_size_bits != 0) {
    reader.SkipBits(preview_size_bits);
  }

  SingleImageManager transform;
  PIK_RETURN_IF_ERROR(PikPassToPixels(dparams, compressed, container, pool,
                                      &reader, io, aux_out, &transform));

  if (dparams.check_decompressed_size &&
      reader.Position() != compressed.size()) {
    return PIK_FAILURE("Pik compressed data size mismatch.");
  }

  io->enc_size = compressed.size();

  return true;
}

}  // namespace pik
