// Copyright 2017 Google Inc. All Rights Reserved.
//
// Use of this source code is governed by an MIT-style
// license that can be found in the LICENSE file or at
// https://opensource.org/licenses/MIT.

#include "pik_frame.h"

#include <limits.h>  // PATH_MAX
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <algorithm>
#include <array>
#include <atomic>
#include <cstdint>
#include <limits>
#include <memory>
#include <string>
#include <vector>
#include "status.h"

#include "ac_strategy.h"
#include "alpha.h"
#include "arch_specific.h"
#include "bilinear_transform.h"
#include "bit_reader.h"
#include "bits.h"
#include "byte_order.h"
#include "color_correlation.h"
#include "color_encoding.h"
#include "common.h"
#include "compiler_specific.h"
#include "compressed_dc.h"
#include "convolve.h"
#include "dct.h"
#include "entropy_coder.h"
#include "external_image.h"
#include "gaborish.h"
#include "gamma_correct.h"
#include "headers.h"
#include "image.h"
#include "lossless16.h"
#include "lossless8.h"
#include "multipass_handler.h"
#include "noise.h"
#include "opsin_image.h"
#include "opsin_inverse.h"
#include "padded_bytes.h"
#include "pik_params.h"
#include "simd/targets.h"
#include "size_coder.h"

namespace pik {
namespace {

/* FrameFlagsFromParams() removed */

void OverrideFlag(const Override o, const uint32_t flag,
                  uint32_t* PIK_RESTRICT flags) {
  if (o == Override::kOn) {
    *flags |= flag;
  } else if (o == Override::kOff) {
    *flags &= ~flag;
  }
}

void OverridePassFlags(const DecompressParams& dparams,
                       FrameHeader* PIK_RESTRICT frame_header) {
  OverrideFlag(dparams.noise, FrameHeader::kNoise, &frame_header->flags);
  OverrideFlag(dparams.gradient, FrameHeader::kGradientMap,
               &frame_header->flags);

  if (dparams.adaptive_reconstruction == Override::kOff) {
    frame_header->have_adaptive_reconstruction = false;
  } else if (dparams.adaptive_reconstruction == Override::kOn) {
    frame_header->have_adaptive_reconstruction = true;
  }
  frame_header->epf_params.use_sharpened = ApplyOverride(
      dparams.epf_use_sharpened, frame_header->epf_params.use_sharpened);
  if (dparams.epf_sigma > 0) {
    frame_header->epf_params.enable_adaptive = false;
    frame_header->epf_params.sigma = dparams.epf_sigma;
  }

  if (dparams.gaborish != -1) {
    frame_header->gaborish = GaborishStrength(dparams.gaborish);
  }
}

void OverrideGroupFlags(const DecompressParams& dparams,
                        const FrameHeader* PIK_RESTRICT frame_header,
                        GroupHeader* PIK_RESTRICT header) {}

// Specializes a 8-bit and 16-bit of rounding from floating point to lossless.
template <typename T>
T RoundForLossless(float in);

template <>
uint8_t RoundForLossless(float in) {
  // NOTE: if in was originally an 8 or 16 bit value, we don't need to round
  // because such values are exactly representable as floats. Rounding is only
  // needed when forcing inexact values back to integers.
  return static_cast<uint8_t>(in + 0.5f);
}

template <>
uint16_t RoundForLossless(float in) {
  return static_cast<uint16_t>(in * 257.0f + 0.5f);
}

// Specializes a 8-bit and 16-bit lossless diff for previous pass.
template <typename T>
T DiffForLossless(float in, float prev);

template <>
uint8_t DiffForLossless(float in, float prev) {
  uint8_t diff = static_cast<int>(RoundForLossless<uint8_t>(in)) -
                 static_cast<int>(RoundForLossless<uint8_t>(prev));
  if (diff > 127)
    diff = (255 - diff) * 2 + 1;
  else
    diff = diff * 2;
  return diff;
}

template <>
uint16_t DiffForLossless(float in, float prev) {
  uint32_t diff = 0xFFFF & (static_cast<int>(RoundForLossless<uint16_t>(in)) -
                            static_cast<int>(RoundForLossless<uint16_t>(prev)));
  if (diff > 32767)
    diff = (65535 - diff) * 2 + 1;
  else
    diff = diff * 2;
  return diff;
}

// Handles one channel c for converting ImageF or Image3F to lossless 8-bit or
// lossless 16-bit, and optionally handles previous pass delta.
template <typename T>
void LosslessChannelPass(const int c, const CodecInOut* io, const Rect& rect,
                         const Image3F& previous_pass, Image<T>* channel_out) {
  size_t xsize = rect.xsize();
  size_t ysize = rect.ysize();
  if (previous_pass.xsize() == 0) {
    for (size_t y = 0; y < ysize; ++y) {
      const float* const PIK_RESTRICT row_in =
          rect.ConstPlaneRow(io->color(), c, y);
      T* const PIK_RESTRICT row_out = channel_out->Row(y);
      for (size_t x = 0; x < xsize; ++x) {
        row_out[x] = RoundForLossless<T>(row_in[x]);
      }
    }
  } else {
    for (size_t y = 0; y < ysize; ++y) {
      const float* const PIK_RESTRICT row_in =
          rect.ConstPlaneRow(io->color(), c, y);
      T* const PIK_RESTRICT row_out = channel_out->Row(y);
      const float* const PIK_RESTRICT row_prev = previous_pass.PlaneRow(0, y);
      for (size_t x = 0; x < xsize; ++x) {
        row_out[x] = DiffForLossless<T>(row_in[x], row_prev[x]);
      }
    }
  }
}

/* PixelsToPikLosslessFrame() removed */

// Returns the target size based on whether bitrate or direct targetsize is
// given.
size_t TargetSize(const CompressParams& cparams, const Rect& rect) {
  if (cparams.target_size > 0) {
    return cparams.target_size;
  }
  if (cparams.target_bitrate > 0.0) {
    return 0.5 + cparams.target_bitrate * rect.xsize() * rect.ysize() / 8;
  }
  return 0;
}

/* PikPassHeuristics() removed */

// Max observed: 1.1M on RGB noise with d0.1.
// 512*512*4*2 = 2M should be enough for 16-bit RGBA images.
using GroupSizeCoder = SizeCoderT<0x150F0E0C>;

}  // namespace

/* PixelsToPikPass() removed */

namespace {

Status ValidateImageDimensions(const FileHeader& file_header,
                               const DecompressParams& dparams) {
  const size_t xsize = file_header.xsize();
  const size_t ysize = file_header.ysize();
  if (xsize == 0 || ysize == 0) {
    return PIK_FAILURE("Empty image.");
  }

  static const size_t kMaxWidth = (1 << 25) - 1;
  if (xsize > kMaxWidth) {
    return PIK_FAILURE("Image too wide.");
  }

  const size_t num_pixels = xsize * ysize;
  if (num_pixels > dparams.max_num_pixels) {
    return PIK_FAILURE("Image too big.");
  }

  return true;
}

// Specializes a 8-bit and 16-bit of converting to float from lossless.
float ToFloatForLossless(uint8_t in) { return static_cast<float>(in); }

float ToFloatForLossless(uint16_t in) { return in * (1.0f / 257); }

// Specializes a 8-bit and 16-bit undo of lossless diff.
float UndiffForLossless(uint8_t in, float prev) {
  uint16_t diff;
  if (in % 2 == 0)
    diff = in / 2;
  else
    diff = 255 - (in / 2);
  uint8_t out = diff + static_cast<int>(RoundForLossless<uint8_t>(prev));
  return ToFloatForLossless(out);
}

float UndiffForLossless(uint16_t in, float prev) {
  uint16_t diff;
  if (in % 2 == 0)
    diff = in / 2;
  else
    diff = 65535 - (in / 2);
  uint16_t out = diff + static_cast<int>(RoundForLossless<uint16_t>(prev));
  return ToFloatForLossless(out);
}

// Handles converting lossless 8-bit or lossless 16-bit, to Image3F, with
// option to give 3x same channel at input for grayscale, and optionally
// handles previous pass delta.
template <typename T>
void LosslessChannelDecodePass(int num_channels, const Image<T>** in,
                               const Rect& rect, const Image3F& previous_pass,
                               Image3F* color) {
  size_t xsize = rect.xsize();
  size_t ysize = rect.ysize();

  for (int c = 0; c < num_channels; c++) {
    if (previous_pass.xsize() == 0) {
      for (size_t y = 0; y < ysize; ++y) {
        const T* const PIK_RESTRICT row_in = in[c]->Row(y);
        float* const PIK_RESTRICT row_out = rect.PlaneRow(color, c, y);
        for (size_t x = 0; x < xsize; ++x) {
          row_out[x] = ToFloatForLossless(row_in[x]);
        }
      }
    } else {
      for (size_t y = 0; y < ysize; ++y) {
        const T* const PIK_RESTRICT row_in = in[c]->Row(y);
        float* const PIK_RESTRICT row_out = rect.PlaneRow(color, c, y);
        const float* const PIK_RESTRICT row_prev =
            previous_pass.ConstPlaneRow(c, y);
        for (size_t x = 0; x < xsize; ++x) {
          row_out[x] = UndiffForLossless(row_in[x], row_prev[x]);
        }
      }
    }
  }

  // Grayscale, copy the channel to the other two output channels
  if (num_channels == 1) {
    for (size_t y = 0; y < ysize; ++y) {
      const float* const PIK_RESTRICT row_0 = rect.PlaneRow(color, 0, y);
      float* const PIK_RESTRICT row_1 = rect.PlaneRow(color, 1, y);
      float* const PIK_RESTRICT row_2 = rect.PlaneRow(color, 2, y);
      for (size_t x = 0; x < xsize; ++x) {
        row_1[x] = row_2[x] = row_0[x];
      }
    }
  }
}

Status PikLosslessFrameToPixels(const PaddedBytes& compressed,
                                const FrameHeader& frame_header,
                                size_t* position, Image3F* color,
                                const Rect& rect,
                                const Image3F& previous_pass) {
  if (frame_header.lossless_grayscale) {
    if (frame_header.lossless_16_bits) {
      ImageU image;
      if (!Grayscale16bit_decompress(compressed, position, &image)) {
        return PIK_FAILURE("Lossless decompression failed");
      }
      if (!SameSize(image, rect)) {
        return PIK_FAILURE("Lossless decompression yielded wrong dimensions.");
      }
      const ImageU* array[1] = {&image};
      LosslessChannelDecodePass(1, array, rect, previous_pass, color);
    } else {
      ImageB image;
      if (!Grayscale8bit_decompress(compressed, position, &image)) {
        return PIK_FAILURE("Lossless decompression failed");
      }
      if (!SameSize(image, rect)) {
        return PIK_FAILURE("Lossless decompression yielded wrong dimensions.");
      }
      const ImageB* array[1] = {&image};
      LosslessChannelDecodePass(1, array, rect, previous_pass, color);
    }
  } else {
    if (frame_header.lossless_16_bits) {
      Image3U image;
      if (!Colorful16bit_decompress(compressed, position, &image)) {
        return PIK_FAILURE("Lossless decompression failed");
      }
      if (!SameSize(image, rect)) {
        return PIK_FAILURE("Lossless decompression yielded wrong dimensions.");
      }
      const ImageU* array[3] = {&image.Plane(0), &image.Plane(1),
                                &image.Plane(2)};
      LosslessChannelDecodePass(3, array, rect, previous_pass, color);
    } else {
      Image3B image;
      if (!Colorful8bit_decompress(compressed, position, &image)) {
        return PIK_FAILURE("Lossless decompression failed");
      }
      if (!SameSize(image, rect)) {
        return PIK_FAILURE("Lossless decompression yielded wrong dimensions.");
      }
      const ImageB* array[3] = {&image.Plane(0), &image.Plane(1),
                                &image.Plane(2)};
      LosslessChannelDecodePass(3, array, rect, previous_pass, color);
    }
  }
  return true;
}

// `reader` is a vector of readers (one per pass). Group headers are only
// present in the first pass, thus the group header of this group is read from
// `reader[0]`.
Status PikGroupToPixels(
    const DecompressParams& dparams, const FileHeader& file_header,
    const FrameHeader* frame_header, const PaddedBytes& compressed,
    const Quantizer& quantizer, const ColorCorrelationMap& full_cmap,
    std::vector<BitReader>* reader, Image3F* PIK_RESTRICT opsin_output,
    ImageU* alpha_output, const CodecContext* context, PikInfo* aux_out,
    FrameDecCache* PIK_RESTRICT frame_dec_cache,
    GroupDecCache* PIK_RESTRICT group_dec_cache,
    MultipassHandler* multipass_handler,
    const ColorEncoding& original_color_encoding, size_t downsampling) {
  const Rect& padded_rect = multipass_handler->PaddedGroupRect();
  const Rect& rect = multipass_handler->GroupRect();
  GroupHeader header;
  header.nonserialized_have_alpha = frame_header->has_alpha;
  PIK_RETURN_IF_ERROR(ReadGroupHeader(&(*reader)[0], &header));
  PIK_RETURN_IF_ERROR((*reader)[0].JumpToByteBoundary());
  OverrideGroupFlags(dparams, frame_header, &header);

  if (frame_header->has_alpha) {
    // TODO(lode): do not fail here based on the metadata
    // original_bytes_per_alpha, it should be allowed to use an efficient
    // encoding in pik which differs from what the original had (or
    // alternatively if they must be the same, there should not be two fields)
    if (header.alpha.bytes_per_alpha !=
        file_header.metadata.transcoded.original_bytes_per_alpha) {
      return PIK_FAILURE("Nonuniform alpha bitdepth is not supported yet.");
    }
    if (file_header.metadata.transcoded.original_bytes_per_alpha == 0) {
      return PIK_FAILURE("Header claims to contain alpha but the depth is 0.");
    }
    PIK_RETURN_IF_ERROR(DecodeAlpha(dparams, header.alpha, alpha_output, rect));
  }

  if (frame_header->encoding == ImageEncoding::kLossless) {
    // Done; we'll decode the entire image in one shot later.
    return true;
  }

  ImageSize opsin_size =
      ImageSize::Make(padded_rect.xsize(), padded_rect.ysize());
  const size_t xsize_blocks = DivCeil<size_t>(opsin_size.xsize, kBlockDim);
  const size_t ysize_blocks = DivCeil<size_t>(opsin_size.ysize, kBlockDim);

  Rect group_in_color_tiles(
      multipass_handler->BlockGroupRect().x0() / kColorTileDimInBlocks,
      multipass_handler->BlockGroupRect().y0() / kColorTileDimInBlocks,
      DivCeil(multipass_handler->BlockGroupRect().xsize(),
              kColorTileDimInBlocks),
      DivCeil(multipass_handler->BlockGroupRect().ysize(),
              kColorTileDimInBlocks));

  NoiseParams noise_params;

  InitializeDecCache(*frame_dec_cache, padded_rect, group_dec_cache);

  if (dparams.max_passes == 0) ZeroFillImage(&group_dec_cache->ac);
  for (size_t i = 0; i < frame_header->num_passes && i < dparams.max_passes;
       i++) {
    auto decode = i == 0 ? &DecodeFromBitstream</*first=*/true>
                         : &DecodeFromBitstream</*first=*/false>;
    if (!decode(*frame_header, header, compressed, &(*reader)[i], padded_rect,
                multipass_handler, xsize_blocks, ysize_blocks, full_cmap,
                group_in_color_tiles, &noise_params, quantizer, frame_dec_cache,
                group_dec_cache, aux_out)) {
      return PIK_FAILURE("Pik decoding failed.");
    }
    if (!(*reader)[i].JumpToByteBoundary()) {
      return PIK_FAILURE("Pik bitstream is corrupted.");
    }
  }

  Rect opsin_rect(padded_rect.x0() / downsampling,
                  padded_rect.y0() / downsampling,
                  DivCeil(padded_rect.xsize(), downsampling),
                  DivCeil(padded_rect.ysize(), downsampling));

  // Note: DecodeFromBitstream already performed dequantization.
  ReconOpsinImage(*frame_header, header, quantizer,
                  multipass_handler->BlockGroupRect(), frame_dec_cache,
                  group_dec_cache, opsin_output, opsin_rect, aux_out,
                  downsampling);

  return true;
}

}  // namespace

Status PikPassToPixels(DecompressParams dparams, const PaddedBytes& compressed,
                       const FileHeader& file_header, ThreadPool* pool,
                       BitReader* reader, CodecInOut* io, PikInfo* aux_out,
                       MultipassManager* multipass_manager) {
  PIK_RETURN_IF_ERROR(ValidateImageDimensions(file_header, dparams));

  io->metadata = file_header.metadata;

  // Used when writing the output file unless DecoderHints overrides it.
  io->SetOriginalBitsPerSample(
      file_header.metadata.transcoded.original_bit_depth);
  io->dec_c_original = file_header.metadata.transcoded.original_color_encoding;
  if (io->dec_c_original.icc.empty()) {
    // Removed by MaybeRemoveProfile; fail unless we successfully restore it.
    PIK_RETURN_IF_ERROR(
        ColorManagement::SetProfileFromFields(&io->dec_c_original));
  }

  const size_t xsize = file_header.xsize();
  const size_t ysize = file_header.ysize();
  const size_t padded_xsize = DivCeil(xsize, kBlockDim) * kBlockDim;
  const size_t padded_ysize = DivCeil(ysize, kBlockDim) * kBlockDim;

  FrameHeader header;
  PIK_RETURN_IF_ERROR(ReadPassHeader(reader, &header));

  PIK_RETURN_IF_ERROR(reader->JumpToByteBoundary());

  // TODO(veluca): add kProgressive.
  if (header.encoding != ImageEncoding::kPasses &&
      header.encoding != ImageEncoding::kLossless) {
    return PIK_FAILURE("Unsupported bitstream");
  }

  OverridePassFlags(dparams, &header);

  size_t downsampling;
  if (dparams.max_downsampling >= 8) {
    downsampling = 8;
    dparams.max_passes = 0;
  } else {
    downsampling = 1;
    for (const auto& downsampling_and_num_passes :
         header.downsampling_factor_to_passes) {
      if (dparams.max_downsampling >= downsampling_and_num_passes.first &&
          dparams.max_passes > downsampling_and_num_passes.second) {
        downsampling = downsampling_and_num_passes.first;
        dparams.max_passes = downsampling_and_num_passes.second + 1;
      }
    }
  }
  if (aux_out != nullptr) {
    aux_out->downsampling = downsampling;
  }

  multipass_manager->StartPass(header);

  ImageU alpha;
  if (header.has_alpha) {
    alpha = ImageU(xsize, ysize);
  }

  const size_t xsize_groups = DivCeil(xsize, kGroupDim);
  const size_t ysize_groups = DivCeil(ysize, kGroupDim);
  const size_t num_groups = xsize_groups * ysize_groups;

  std::vector<PikInfo> aux_outs;
  if (aux_out != nullptr) {
    aux_outs.resize(num_groups, *aux_out);
  }
  std::vector<MultipassHandler*> handlers(num_groups);
  {
    for (size_t group_index = 0; group_index < num_groups; ++group_index) {
      const size_t gx = group_index % xsize_groups;
      const size_t gy = group_index / xsize_groups;
      const size_t x = gx * kGroupDim;
      const size_t y = gy * kGroupDim;
      Rect rect(x, y, kGroupDim, kGroupDim, xsize, ysize);
      handlers[group_index] =
          multipass_manager->GetGroupHandler(group_index, rect);
    }
  }

  const size_t xsize_blocks = padded_xsize / kBlockDim;
  const size_t ysize_blocks = padded_ysize / kBlockDim;

  FrameDecCache frame_dec_cache;
  frame_dec_cache.use_new_dc = dparams.use_new_dc;
  frame_dec_cache.grayscale = header.flags & FrameHeader::kGrayscaleOpt;
  frame_dec_cache.ac_strategy = AcStrategyImage(xsize_blocks, ysize_blocks);
  frame_dec_cache.raw_quant_field = ImageI(xsize_blocks, ysize_blocks);
  frame_dec_cache.ar_sigma_lut_ids = ImageB(xsize_blocks, ysize_blocks);
  frame_dec_cache.dequant_control_field =
      ImageB(DivCeil(xsize, kTileDim), DivCeil(ysize, kTileDim));

  ColorCorrelationMap cmap(xsize, ysize);

  // TODO(veluca): deserialize quantization tables from the bitstream.

  Quantizer quantizer(&frame_dec_cache.matrices, 0, 0);
  BlockDictionary block_dictionary;

  std::vector<GroupDecCache> group_dec_caches(NumThreads(pool));

  if (header.encoding == ImageEncoding::kPasses) {
    PIK_RETURN_IF_ERROR(frame_dec_cache.matrices.Decode(reader));
    PIK_RETURN_IF_ERROR(quantizer.Decode(reader));
    PIK_RETURN_IF_ERROR(reader->JumpToByteBoundary());

    // TODO(veluca): decode quantization table mapping.

    DecodeColorMap(reader, &cmap.ytob_map, &cmap.ytob_dc);
    DecodeColorMap(reader, &cmap.ytox_map, &cmap.ytox_dc);
    PIK_RETURN_IF_ERROR(DecodeDCGroups(
        reader, compressed, header, xsize_blocks, ysize_blocks, quantizer, cmap,
        pool, multipass_manager, &frame_dec_cache, &group_dec_caches, aux_out));
    PIK_RETURN_IF_ERROR(
        block_dictionary.Decode(reader, padded_xsize, padded_ysize));

    // TODO(veluca): think of splitting this in DC groups.
    PIK_RETURN_IF_ERROR(DecodeDequantControlField(
        reader, &frame_dec_cache.dequant_control_field));
    PIK_RETURN_IF_ERROR(DecodeDequantControlFieldMap(
        reader, frame_dec_cache.raw_quant_field,
        frame_dec_cache.dequant_control_field, frame_dec_cache.dequant_map));
    multipass_manager->SaveAcStrategy(frame_dec_cache.ac_strategy);
    multipass_manager->SaveQuantField(frame_dec_cache.raw_quant_field);
  }

  Image3F opsin(DivCeil(padded_xsize, downsampling),
                DivCeil(padded_ysize, downsampling));

  // Read TOCs.
  std::vector<std::vector<size_t>> group_offsets(header.num_passes);
  std::vector<size_t> group_codes_begin(header.num_passes);
  for (size_t i = 0; i < header.num_passes; i++) {
    std::vector<size_t>& group_offsets_pass = group_offsets[i];
    group_offsets_pass.reserve(num_groups + 1);
    group_offsets_pass.push_back(0);
    for (size_t group_index = 0; group_index < num_groups; ++group_index) {
      const uint32_t size = GroupSizeCoder::Decode(reader);
      group_offsets_pass.push_back(group_offsets_pass.back() + size);
    }
    PIK_RETURN_IF_ERROR(reader->JumpToByteBoundary());
    // On first pass, read lossless.
    if (header.encoding == ImageEncoding::kLossless && i == 0) {
      const Rect rect(0, 0, xsize, ysize);

      size_t pos = reader->Position();
      const size_t before_pos = pos;

      Image3F previous_pass;
      PIK_RETURN_IF_ERROR(multipass_manager->GetPreviousPass(
          io->dec_c_original, pool, &previous_pass));
      PIK_RETURN_IF_ERROR(PikLosslessFrameToPixels(
          compressed, header, &pos, &opsin, rect, previous_pass));
      reader->SkipBits((pos - before_pos) * kBitsPerByte);
      // Byte-wise; no need to jump to boundary.
    }
    // Pretend all groups of this pass are read.
    group_codes_begin[i] = reader->Position();
    reader->SkipBits(group_offsets_pass.back() * kBitsPerByte);
    if (reader->Position() > compressed.size()) {
      return PIK_FAILURE("Group code extends after stream end");
    }
  }

  // Decode groups.
  std::atomic<int> num_errors{0};
  const auto process_group = [&](const int group_index, const int thread) {
    std::vector<BitReader> readers;
    for (size_t i = 0; i < header.num_passes; i++) {
      size_t group_code_offset = group_offsets[i][group_index];
      size_t group_reader_limit = group_offsets[i][group_index + 1];
      // TODO(user): this looks ugly; we should get rid of PaddedBytes
      //               parameter once it is wrapped into BitReader; otherwise
      //               it is easy to screw the things up.
      readers.emplace_back(compressed.data(),
                           group_codes_begin[i] + group_reader_limit);
      readers.back().SkipBits((group_codes_begin[i] + group_code_offset) *
                              kBitsPerByte);
    }

    PikInfo* my_aux_out = aux_out ? &aux_outs[group_index] : nullptr;
    if (!PikGroupToPixels(dparams, file_header, &header, compressed, quantizer,
                          cmap, &readers, &opsin, &alpha, io->Context(),
                          my_aux_out, &frame_dec_cache,
                          &group_dec_caches[thread], handlers[group_index],
                          io->dec_c_original, downsampling)) {
      num_errors.fetch_add(1);
      return;
    }
  };
  {
    RunOnPool(pool, 0, num_groups, process_group, "PikPassToPixels");
  }

  PIK_RETURN_IF_ERROR(num_errors.load(std::memory_order_relaxed) == 0);

  if (aux_out != nullptr) {
    for (size_t group_index = 0; group_index < num_groups; ++group_index) {
      aux_out->Assimilate(aux_outs[group_index]);
    }
  }

  if (header.encoding == ImageEncoding::kPasses) {
    multipass_manager->RestoreOpsin(&opsin);
    multipass_manager->SetDecodedPass(opsin);

    PIK_RETURN_IF_ERROR(FinalizeFrameDecoding(
        &opsin, file_header.xsize(), file_header.ysize(), header, NoiseParams(),
        quantizer, block_dictionary, pool, &frame_dec_cache, aux_out,
        downsampling));
    // From now on, `opsin` is actually linear sRGB.

    if (header.flags & FrameHeader::kGrayscaleOpt) {
      // Force all channels to gray
      for (size_t y = 0; y < opsin.ysize(); ++y) {
        float* PIK_RESTRICT row_r = opsin.PlaneRow(0, y);
        float* PIK_RESTRICT row_g = opsin.PlaneRow(1, y);
        float* PIK_RESTRICT row_b = opsin.PlaneRow(2, y);
        for (size_t x = 0; x < opsin.xsize(); x++) {
          float gray = row_r[x] * 0.299 + row_g[x] * 0.587 + row_b[x] * 0.114;
          row_r[x] = row_g[x] = row_b[x] = gray;
        }
      }
    }
    const ColorEncoding& c =
        io->Context()->c_linear_srgb[io->dec_c_original.IsGray()];
    io->SetFromImage(std::move(opsin), c);
  } else if (header.encoding == ImageEncoding::kLossless) {
    io->SetFromImage(std::move(opsin), io->dec_c_original);
    io->ShrinkTo(xsize, ysize);
    multipass_manager->SetDecodedPass(io);
  } else {
    return PIK_FAILURE("Unsupported image encoding");
  }

  if (header.has_alpha) {
    io->SetAlpha(std::move(alpha),
                 8 * file_header.metadata.transcoded.original_bytes_per_alpha);
  }

  io->ShrinkTo(DivCeil(xsize, downsampling), DivCeil(ysize, downsampling));

  return true;
}

}  // namespace pik
