// Copyright 2017 Google Inc. All Rights Reserved.
//
// Use of this source code is governed by an MIT-style
// license that can be found in the LICENSE file or at
// https://opensource.org/licenses/MIT.

#ifndef PIK_PIK_INFO_H_
#define PIK_PIK_INFO_H_

// Optional output information for debugging and analyzing size usage.

#include <cstddef>
#include <sstream>
#include <string>
#include <vector>
#include "adaptive_reconstruction_fwd.h"
#include "codec.h"
#include "image.h"
#include "pik_inspection.h"

namespace pik {

struct PikImageSizeInfo {
  PikImageSizeInfo() {}

  void Assimilate(const PikImageSizeInfo& victim) {
    num_clustered_histograms += victim.num_clustered_histograms;
    histogram_size += victim.histogram_size;
    entropy_coded_bits += victim.entropy_coded_bits;
    extra_bits += victim.extra_bits;
    total_size += victim.total_size;
    clustered_entropy += victim.clustered_entropy;
  }
  /* Print() removed */
  size_t num_clustered_histograms = 0;
  size_t histogram_size = 0;
  size_t entropy_coded_bits = 0;
  size_t extra_bits = 0;
  size_t total_size = 0;
  double clustered_entropy = 0.0f;
};

enum {
  kLayerHeader = 0,
  kLayerQuant,
  kLayerDequantTables,
  kLayerOrder,
  kLayerDC,
  kLayerCmap,
  kLayerControlFields,
  kLayerAC,
  kLayerDictionary,
  kNumImageLayers
};
/* kImageLayers[] removed */

struct TestingAux {
  Image3F* ac_prediction = nullptr;
};

// Metadata and statistics gathered during compression or decompression.
struct PikInfo {
  PikInfo() : layers(kNumImageLayers) {}

  PikInfo(const PikInfo&) = default;

  void Assimilate(const PikInfo& victim) {
    for (int i = 0; i < layers.size(); ++i) {
      layers[i].Assimilate(victim.layers[i]);
    }
    num_blocks += victim.num_blocks;
    num_dct2_blocks += victim.num_dct2_blocks;
    num_dct4_blocks += victim.num_dct4_blocks;
    num_dct16_blocks += victim.num_dct16_blocks;
    num_dct32_blocks += victim.num_dct32_blocks;
    entropy_estimate += victim.entropy_estimate;
    num_butteraugli_iters += victim.num_butteraugli_iters;
    /* stats removed */
    adaptive_reconstruction_aux.Assimilate(victim.adaptive_reconstruction_aux);
  }

  PikImageSizeInfo TotalImageSize() const {
    PikImageSizeInfo total;
    for (int i = 0; i < layers.size(); ++i) {
      total.Assimilate(layers[i]);
    }
    return total;
  }

  /* Print() removed */

  /* DumpImage() removed */
  /* DumpImage() removed */

  /* DumpCoeffImage() removed */

  void SetInspectorImage3F(pik::InspectorImage3F inspector) {
    inspector_image3f_ = inspector;
  }

  // Allows hooking intermediate data inspection into various
  // places of the PIK processing pipeline. Returns true iff
  // processing should proceed.
  bool InspectImage3F(const char* label, const Image3F& image) {
    if (inspector_image3f_ != nullptr) {
      return inspector_image3f_(label, image);
    }
    return true;
  }

  std::vector<PikImageSizeInfo> layers;
  size_t num_blocks = 0;
  // Number of blocks that use larger DCT. Only set in the encoder.
  size_t num_dct2_blocks = 0;
  size_t num_dct4_blocks = 0;
  size_t num_dct16_blocks = 0;
  size_t num_dct32_blocks = 0;
  // Estimate of compressed size according to entropy-given lower bounds.
  float entropy_estimate = 0;
  int num_butteraugli_iters = 0;
  // If not empty, additional debugging information (e.g. debug images) is
  // saved in files with this prefix.
  std::string debug_prefix;

  // By how much the decoded image was downsampled relative to the encoded
  // image.
  size_t downsampling = 1;

  AdaptiveReconstructionAux adaptive_reconstruction_aux;
  /* stats removed */

  pik::InspectorImage3F inspector_image3f_;

  // WARNING: this is actually an INPUT to some code, and must be
  // copy-initialized from aux_out to aux_outs.
  TestingAux testing_aux;
};

/* WantDebugOutput() removed */

}  // namespace pik

#endif  // PIK_PIK_INFO_H_
