// Copyright 2017 Google Inc. All Rights Reserved.
//
// Use of this source code is governed by an MIT-style
// license that can be found in the LICENSE file or at
// https://opensource.org/licenses/MIT.

#ifndef PIK_SIMD_SCALAR_H_
#define PIK_SIMD_SCALAR_H_

// Single-element vectors and operations.

#include "compiler_specific.h"
#include "shared.h"
#include "targets.h"
#include "util.h"

namespace pik {

// Never override NONE.
template <>
struct PartTargetT<1, NONE> {
  using type = NONE;
};

// Shorthand for a scalar; note that scalar<T> is the actual data class.
template <typename T>
using Scalar = Desc<T, 1, NONE>;

// Returned by set_shift_*_count; do not use directly.
struct scalar_shift_left_count {
  int count;
};
struct scalar_shift_right_count {
  int count;
};

// (Wrapper class required for overloading comparison operators.)
template <typename T>
struct scalar {
  SIMD_INLINE scalar() {}
  scalar(const scalar&) = default;
  scalar& operator=(const scalar&) = default;
  SIMD_INLINE explicit scalar(const T t) : raw(t) {}

  SIMD_INLINE scalar& operator*=(const scalar other) {
    return *this = (*this * other);
  }
  SIMD_INLINE scalar& operator/=(const scalar other) {
    return *this = (*this / other);
  }
  SIMD_INLINE scalar& operator+=(const scalar other) {
    return *this = (*this + other);
  }
  SIMD_INLINE scalar& operator-=(const scalar other) {
    return *this = (*this - other);
  }
  SIMD_INLINE scalar& operator&=(const scalar other) {
    return *this = (*this & other);
  }
  SIMD_INLINE scalar& operator|=(const scalar other) {
    return *this = (*this | other);
  }
  SIMD_INLINE scalar& operator^=(const scalar other) {
    return *this = (*this ^ other);
  }

  T raw;
};

template <typename T>
struct VecT<T, 1, NONE> {
  using type = scalar<T>;
};

// ------------------------------ Cast

template <typename T, typename FromT>
SIMD_INLINE scalar<T> cast_to(Scalar<T>, scalar<FromT> v) {
  static_assert(sizeof(T) <= sizeof(FromT), "Promoting is undefined");
  T to;
  CopyBytes<sizeof(FromT)>(&v.raw, &to);
  return scalar<T>(to);
}

// ------------------------------ Set

template <typename T>
SIMD_INLINE scalar<T> setzero(Scalar<T>) {
  return scalar<T>(T(0));
}

template <typename T, typename T2>
SIMD_INLINE scalar<T> set1(Scalar<T>, const T2 t) {
  return scalar<T>(t);
}

/* iota() removed */

template <typename T>
SIMD_INLINE scalar<T> undefined(Scalar<T>) {
  return scalar<T>(0);
}

// ================================================== SHIFTS

// ------------------------------ Shift lanes by constant #bits

template <int kBits, typename T>
SIMD_INLINE scalar<T> shift_left(const scalar<T> v) {
  static_assert(0 <= kBits && kBits < sizeof(T) * 8, "Invalid shift");
  return scalar<T>(v.raw << kBits);
}

template <int kBits, typename T>
SIMD_INLINE scalar<T> shift_right(const scalar<T> v) {
  static_assert(0 <= kBits && kBits < sizeof(T) * 8, "Invalid shift");
  return scalar<T>(v.raw >> kBits);
}

// ------------------------------ Shift lanes by same variable #bits

template <typename T>
SIMD_INLINE scalar_shift_left_count set_shift_left_count(Scalar<T>,
                                                         const int bits) {
  return scalar_shift_left_count{bits};
}

template <typename T>
SIMD_INLINE scalar_shift_right_count set_shift_right_count(Scalar<T>,
                                                           const int bits) {
  return scalar_shift_right_count{bits};
}

/* shift_left_same() removed */
/* shift_right_same() removed */

// ------------------------------ Shift lanes by independent variable #bits

// Single-lane => same as above except for the argument type.
template <typename T>
SIMD_INLINE scalar<T> operator<<(const scalar<T> v, const scalar<T> bits) {
  return scalar<T>(v.raw << bits.raw);
}
template <typename T>
SIMD_INLINE scalar<T> operator>>(const scalar<T> v, const scalar<T> bits) {
  return scalar<T>(v.raw >> bits.raw);
}

// ================================================== LOGICAL

template <typename Bits>
struct BitwiseOp {
  template <typename T, class Op>
  scalar<T> operator()(const scalar<T> a, const scalar<T> b,
                       const Op& op) const {
    static_assert(sizeof(T) == sizeof(Bits), "Float/int size mismatch");
    Bits ia, ib;
    CopyBytes<sizeof(Bits)>(&a, &ia);
    CopyBytes<sizeof(Bits)>(&b, &ib);
    ia = op(ia, ib);
    T ret;
    CopyBytes<sizeof(Bits)>(&ia, &ret);
    return scalar<T>(ret);
  }
};

// ------------------------------ Bitwise AND

template <typename T>
SIMD_INLINE scalar<T> operator&(const scalar<T> a, const scalar<T> b) {
  return scalar<T>(a.raw & b.raw);
}
template <>
SIMD_INLINE scalar<float> operator&(const scalar<float> a,
                                    const scalar<float> b) {
  return BitwiseOp<int32_t>()(a, b, [](int32_t i, int32_t j) { return i & j; });
}
template <>
SIMD_INLINE scalar<double> operator&(const scalar<double> a,
                                     const scalar<double> b) {
  return BitwiseOp<int64_t>()(a, b, [](int64_t i, int64_t j) { return i & j; });
}

// ------------------------------ Bitwise AND-NOT

// Returns ~a & b.
template <typename T>
SIMD_INLINE scalar<T> andnot(const scalar<T> a, const scalar<T> b) {
  return scalar<T>(~a.raw & b.raw);
}
template <>
SIMD_INLINE scalar<float> andnot(const scalar<float> a, const scalar<float> b) {
  return BitwiseOp<int32_t>()(a, b,
                              [](int32_t i, int32_t j) { return ~i & j; });
}

/* andnot() removed */

// ------------------------------ Bitwise OR

template <typename T>
SIMD_INLINE scalar<T> operator|(const scalar<T> a, const scalar<T> b) {
  return scalar<T>(a.raw | b.raw);
}
template <>
SIMD_INLINE scalar<float> operator|(const scalar<float> a,
                                    const scalar<float> b) {
  return BitwiseOp<int32_t>()(a, b, [](int32_t i, int32_t j) { return i | j; });
}
template <>
SIMD_INLINE scalar<double> operator|(const scalar<double> a,
                                     const scalar<double> b) {
  return BitwiseOp<int64_t>()(a, b, [](int64_t i, int64_t j) { return i | j; });
}

// ------------------------------ Bitwise XOR

template <typename T>
SIMD_INLINE scalar<T> operator^(const scalar<T> a, const scalar<T> b) {
  return scalar<T>(a.raw ^ b.raw);
}
template <>
SIMD_INLINE scalar<float> operator^(const scalar<float> a,
                                    const scalar<float> b) {
  return BitwiseOp<int32_t>()(a, b, [](int32_t i, int32_t j) { return i ^ j; });
}
template <>
SIMD_INLINE scalar<double> operator^(const scalar<double> a,
                                     const scalar<double> b) {
  return BitwiseOp<int64_t>()(a, b, [](int64_t i, int64_t j) { return i ^ j; });
}

// ------------------------------ Select/blend

// Returns a mask for use by select().
SIMD_INLINE scalar<float> condition_from_sign(const scalar<float> v) {
  const Scalar<float> df;
  const Scalar<int32_t> di;
  return cast_to(df, shift_right<31>(cast_to(di, v)));
}
SIMD_INLINE scalar<double> condition_from_sign(const scalar<double> v) {
  const Scalar<double> df;
  const Scalar<int64_t> di;
  return cast_to(df, shift_right<63>(cast_to(di, v)));
}

// Returns mask ? b : a. "mask" must either have been returned by
// selector_from_mask, or callers must ensure its lanes are T(0) or ~T(0).
template <typename T>
SIMD_INLINE scalar<T> select(const scalar<T> a, const scalar<T> b,
                             const scalar<T> mask) {
  return (mask & b) | andnot(mask, a);
}

// ================================================== ARITHMETIC

template <typename T>
SIMD_INLINE scalar<T> operator+(const scalar<T> a, const scalar<T> b) {
  const uint64_t a64 = static_cast<int64_t>(a.raw);
  const uint64_t b64 = static_cast<int64_t>(b.raw);
  return scalar<T>((a64 + b64) & ~T(0));
}
SIMD_INLINE scalar<float> operator+(const scalar<float> a,
                                    const scalar<float> b) {
  return scalar<float>(a.raw + b.raw);
}
SIMD_INLINE scalar<double> operator+(const scalar<double> a,
                                     const scalar<double> b) {
  return scalar<double>(a.raw + b.raw);
}

template <typename T>
SIMD_INLINE scalar<T> operator-(const scalar<T> a, const scalar<T> b) {
  const uint64_t a64 = static_cast<int64_t>(a.raw);
  const uint64_t b64 = static_cast<int64_t>(b.raw);
  return scalar<T>((a64 - b64) & ~T(0));
}
SIMD_INLINE scalar<float> operator-(const scalar<float> a,
                                    const scalar<float> b) {
  return scalar<float>(a.raw - b.raw);
}
SIMD_INLINE scalar<double> operator-(const scalar<double> a,
                                     const scalar<double> b) {
  return scalar<double>(a.raw - b.raw);
}

// ------------------------------ Saturating addition

// Returns a + b clamped to the destination range.

// Unsigned
SIMD_INLINE scalar<uint8_t> saturated_add(const scalar<uint8_t> a,
                                          const scalar<uint8_t> b) {
  return scalar<uint8_t>(SIMD_MIN(SIMD_MAX(0, a.raw + b.raw), 255));
}
/* saturated_add(const scalar<uint16_t> ...) removed */

// Signed
/* saturated_add(const scalar<int8_t> ...) removed */
/* saturated_add(const scalar<int16_t> ...) removed */

// ------------------------------ Saturating subtraction

// Returns a - b clamped to the destination range.

// Unsigned
SIMD_INLINE scalar<uint8_t> saturated_subtract(const scalar<uint8_t> a,
                                               const scalar<uint8_t> b) {
  return scalar<uint8_t>(SIMD_MIN(SIMD_MAX(0, a.raw - b.raw), 255));
}
/* saturated_subtract(const scalar<uint16_t> ...) removed */

// Signed
/* saturated_subtract(const scalar<int8_t> ...) removed */
/* saturated_subtract(const scalar<int16_t> ...) removed */

// ------------------------------ Average

/* average_round(const scalar<uint8_t> ...) removed */
/* average_round(const scalar<uint16_t> ...) removed */

// ------------------------------ Absolute value

template <typename T>
SIMD_INLINE scalar<T> abs(const scalar<T> a) {
  const T i = a.raw;
  return (i >= 0 || i == LimitsMin<T>()) ? a : scalar<T>(-i);
}

// ------------------------------ min/max

template <typename T>
SIMD_INLINE scalar<T> min(const scalar<T> a, const scalar<T> b) {
  return scalar<T>(SIMD_MIN(a.raw, b.raw));
}

template <typename T>
SIMD_INLINE scalar<T> max(const scalar<T> a, const scalar<T> b) {
  return scalar<T>(SIMD_MAX(a.raw, b.raw));
}

// Returns the closest value to v within [lo, hi].
template <typename T>
SIMD_INLINE scalar<T> clamp(const scalar<T> v, const scalar<T> lo,
                            const scalar<T> hi) {
  return min(max(lo, v), hi);
}

// ------------------------------ Floating-point negate

/* neg(const scalar<float>) removed */

/* neg(const scalar<double>) removed */

// ------------------------------ mul/div

template <typename T>
SIMD_INLINE scalar<T> operator*(const scalar<T> a, const scalar<T> b) {
  if (IsFloat<T>()) {
    return scalar<T>(static_cast<T>(double(a.raw) * b.raw));
  } else if (IsSigned<T>()) {
    return scalar<T>(static_cast<T>(int64_t(a.raw) * b.raw));
  } else {
    return scalar<T>(static_cast<T>(uint64_t(a.raw) * b.raw));
  }
}

template <typename T>
SIMD_INLINE scalar<T> operator/(const scalar<T> a, const scalar<T> b) {
  return scalar<T>(a.raw / b.raw);
}

// "Extensions": useful but not quite performance-portable operations. We add
// functions to this namespace in multiple places.
namespace ext {

// Returns the upper 16 bits of a * b in each lane.
SIMD_INLINE scalar<int16_t> mul_high(const scalar<int16_t> a,
                                     const scalar<int16_t> b) {
  return scalar<int16_t>((a.raw * b.raw) >> 16);
}

}  // namespace ext

/* mul_high_round() removed */

/* mul_even(const scalar<int32_t> ...) removed */
/* mul_even(const scalar<uint32_t> ...) removed */

// Approximate reciprocal
SIMD_INLINE scalar<float> approximate_reciprocal(const scalar<float> v) {
  return scalar<float>(1.0f / v.raw);
}

// ------------------------------ Floating-point multiply-add variants

template <typename T>
SIMD_INLINE scalar<T> mul_add(const scalar<T> mul, const scalar<T> x,
                            const scalar<T> add) {
  return mul * x + add;
}

template <typename T>
SIMD_INLINE scalar<T> nmul_add(const scalar<T> mul, const scalar<T> x,
                               const scalar<T> add) {
  return add - mul * x;
}

/* fadd() removed */

/* fsub() removed */

/* fnadd() removed */

// Slightly more expensive on ARM (extra negate)
namespace ext {

template <typename T>
SIMD_INLINE scalar<T> mul_subtract(const scalar<T> mul, const scalar<T> x,
                                   const scalar<T> sub) {
  return mul * x - sub;
}

template <typename T>
SIMD_INLINE scalar<T> nmul_subtract(const scalar<T> mul, const scalar<T> x,
                                    const scalar<T> sub) {
  return neg(mul) * x - sub;
}

}  // namespace ext

// ------------------------------ Floating-point square root

/* approximate_reciprocal_sqrt() removed */

/* sqrt(const scalar<float> removed */
/* sqrt(const scalar<double> removed */

// ------------------------------ Floating-point rounding

// Approximation of round-to-nearest for numbers representable as integers.
SIMD_INLINE scalar<float> round(const scalar<float> v) {
  const float bias = v.raw < 0.0f ? -0.5f : 0.5f;
  return scalar<float>(static_cast<int32_t>(v.raw + bias));
}
SIMD_INLINE scalar<double> round(const scalar<double> v) {
  const double bias = v.raw < 0.0 ? -0.5 : 0.5;
  return scalar<double>(static_cast<int64_t>(v.raw + bias));
}

/* trunc(const scalar<float>) removed */
/* trunc(const scalar<double>) removed */

template <typename Float, typename Bits, int kMantissaBits, int kExponentBits,
          class V>
V Ceiling(const V v) {
  const Bits kExponentMask = (1ull << kExponentBits) - 1;
  const Bits kMantissaMask = (1ull << kMantissaBits) - 1;
  const Bits kBias = kExponentMask / 2;

  Float f = v.raw;
  const bool positive = f > 0.0f;

  Bits bits;
  CopyBytes<sizeof(Bits)>(&v, &bits);

  const int exponent = ((bits >> kMantissaBits) & kExponentMask) - kBias;
  // Already an integer.
  if (exponent >= kMantissaBits) return v;
  // |v| <= 1 => 0 or 1.
  if (exponent < 0) return V(positive);

  const Bits mantissa_mask = kMantissaMask >> exponent;
  // Already an integer
  if ((bits & mantissa_mask) == 0) return v;

  // Clear fractional bits and round up
  if (positive) bits += (kMantissaMask + 1) >> exponent;
  bits &= ~mantissa_mask;

  CopyBytes<sizeof(Bits)>(&bits, &f);
  return V(f);
}

template <typename Float, typename Bits, int kMantissaBits, int kExponentBits,
          class V>
V Floor(const V v) {
  const Bits kExponentMask = (1ull << kExponentBits) - 1;
  const Bits kMantissaMask = (1ull << kMantissaBits) - 1;
  const Bits kBias = kExponentMask / 2;

  Float f = v.raw;
  const bool negative = f < 0.0f;

  Bits bits;
  CopyBytes<sizeof(Bits)>(&v, &bits);

  const int exponent = ((bits >> kMantissaBits) & kExponentMask) - kBias;
  // Already an integer.
  if (exponent >= kMantissaBits) return v;
  // |v| <= 1 => -1 or 0.
  if (exponent < 0) return V(negative ? -1.0 : 0.0f);

  const Bits mantissa_mask = kMantissaMask >> exponent;
  // Already an integer
  if ((bits & mantissa_mask) == 0) return v;

  // Clear fractional bits and round down
  if (negative) bits += (kMantissaMask + 1) >> exponent;
  bits &= ~mantissa_mask;

  CopyBytes<sizeof(Bits)>(&bits, &f);
  return V(f);
}

/* ceil(const scalar<float>) removed */
/* ceil(const scalar<double>) removed */

// Toward -infinity, aka floor
SIMD_INLINE scalar<float> floor(const scalar<float> v) {
  return Floor<float, uint32_t, 23, 8>(v);
}
/* floor(const scalar<double>) removed */

// ================================================== COMPARE

// Comparisons fill a lane with 1-bits if the condition is true, else 0.
template <typename T>
scalar<T> ComparisonResult(const bool result) {
  T ret;
  SetBytes(result ? 0xFF : 0, &ret);
  return scalar<T>(ret);
}

template <typename T>
SIMD_INLINE scalar<T> operator==(const scalar<T> a, const scalar<T> b) {
  return ComparisonResult<T>(a.raw == b.raw);
}

template <typename T>
SIMD_INLINE scalar<T> operator<(const scalar<T> a, const scalar<T> b) {
  return ComparisonResult<T>(a.raw < b.raw);
}
template <typename T>
SIMD_INLINE scalar<T> operator>(const scalar<T> a, const scalar<T> b) {
  return ComparisonResult<T>(a.raw > b.raw);
}

template <typename T>
SIMD_INLINE scalar<T> operator<=(const scalar<T> a, const scalar<T> b) {
  return ComparisonResult<T>(a.raw <= b.raw);
}
template <typename T>
SIMD_INLINE scalar<T> operator>=(const scalar<T> a, const scalar<T> b) {
  return ComparisonResult<T>(a.raw >= b.raw);
}

// ================================================== MEMORY

// ------------------------------ Load

template <typename T>
SIMD_INLINE scalar<T> load(Scalar<T>, const T* SIMD_RESTRICT aligned) {
  T t;
  CopyBytes<sizeof(T)>(aligned, &t);
  return scalar<T>(t);
}

template <typename T>
SIMD_INLINE scalar<T> load_unaligned(Scalar<T> d, const T* SIMD_RESTRICT p) {
  return load(d, p);
}

// In some use cases, "load single lane" is sufficient; otherwise avoid this.
template <typename T>
SIMD_INLINE scalar<T> load_dup128(Scalar<T> d, const T* SIMD_RESTRICT aligned) {
  return load(d, aligned);
}

// ------------------------------ Store

template <typename T>
SIMD_INLINE void store(const scalar<T> v, Scalar<T>, T* SIMD_RESTRICT aligned) {
  CopyBytes<sizeof(T)>(&v.raw, aligned);
}

template <typename T>
SIMD_INLINE void store_unaligned(const scalar<T> v, Scalar<T> d,
                                 T* SIMD_RESTRICT p) {
  return store(v, d, p);
}

// ------------------------------ "Non-temporal" stores

/* stream() removed */

// ------------------------------ Gather

// "Extensions": useful but not quite performance-portable operations. We add
// functions to this namespace in multiple places.
namespace ext {

template <typename T, typename Offset>
SIMD_INLINE scalar<T> gather_offset(Scalar<T> d, const T* base,
                                    const scalar<Offset> offset) {
  static_assert(sizeof(T) == sizeof(Offset), "SVE requires same size base/ofs");
  char sizeof_t[sizeof(T)];
  const uintptr_t addr = reinterpret_cast<uintptr_t>(base) + offset.raw;
  return load(d, reinterpret_cast<const T*>(addr));
}

template <typename T, typename Index>
SIMD_INLINE scalar<T> gather_index(Scalar<T> d, const T* SIMD_RESTRICT base,
                                   const scalar<Index> index) {
  static_assert(sizeof(T) == sizeof(Index), "SVE requires same size base/idx");
  return load(d, base + index.raw);
}

}  // namespace ext

// ================================================== CONVERT

template <typename FromT, typename ToT>
SIMD_INLINE scalar<ToT> convert_to(Desc<ToT, 1, NONE>,
                                   const scalar<FromT> from) {
  return scalar<ToT>(from.raw);
}

/* convert_to(Scalar<float> ...) removed */

/* convert_to(Scalar<int32_t> ...) removed */

SIMD_INLINE scalar<uint32_t> u32_from_u8(const scalar<uint8_t> v) {
  return convert_to(Scalar<uint32_t>(), v);
}

SIMD_INLINE scalar<uint8_t> u8_from_u32(const scalar<uint32_t> v) {
  return convert_to(Scalar<uint8_t>(), v);
}

/* nearest_int() removed */

// ================================================== SWIZZLE

// Unsupported: shift_*_bytes, combine_shift_right_bytes, interleave_*,
// other_half, shuffle_*, sums_of_u8x8, sum_of_lanes - these require more than
// one lane and/or actual 128-bit vectors.

// ------------------------------ Broadcast/splat any lane

/* broadcast() removed */

// ------------------------------ Zip/unpack

SIMD_INLINE scalar<uint16_t> zip_lo(const scalar<uint8_t> a,
                                  const scalar<uint8_t> b) {
  return scalar<uint16_t>((uint32_t(b.raw) << 8) + a.raw);
}
SIMD_INLINE scalar<uint32_t> zip_lo(const scalar<uint16_t> a,
                                  const scalar<uint16_t> b) {
  return scalar<uint32_t>((uint32_t(b.raw) << 16) + a.raw);
}
SIMD_INLINE scalar<uint64_t> zip_lo(const scalar<uint32_t> a,
                                  const scalar<uint32_t> b) {
  return scalar<uint64_t>((uint64_t(b.raw) << 32) + a.raw);
}
SIMD_INLINE scalar<int16_t> zip_lo(const scalar<int8_t> a,
                                   const scalar<int8_t> b) {
  return scalar<int16_t>((uint32_t(b.raw) << 8) + a.raw);
}
SIMD_INLINE scalar<int32_t> zip_lo(const scalar<int16_t> a,
                                   const scalar<int16_t> b) {
  return scalar<int32_t>((uint32_t(b.raw) << 16) + a.raw);
}
SIMD_INLINE scalar<int64_t> zip_lo(const scalar<int32_t> a,
                                   const scalar<int32_t> b) {
  return scalar<int64_t>((uint64_t(b.raw) << 32) + a.raw);
}

template <typename T>
SIMD_INLINE auto zip_hi(const scalar<T> a, const scalar<T> b)
    -> decltype(zip_lo(a, b)) {
  return zip_lo(a, b);
}

// ------------------------------ Parts

template <typename T, typename T2>
SIMD_INLINE scalar<T> set_part(Scalar<T>, const T2 t) {
  return scalar<T>(t);
}

template <typename T>
SIMD_INLINE T get_part(Scalar<T>, const scalar<T> v) {
  return v.raw;
}

template <typename T>
SIMD_INLINE scalar<T> any_part(Scalar<T>, const scalar<T> v) {
  return v;
}

/* broadcast_part() removed */

// ================================================== MISC

// "Extensions": useful but not quite performance-portable operations. We add
// functions to this namespace in multiple places.
namespace ext {

// Returns a bit array of the most significant bit of each byte in "v", i.e.
// sum_i=0..15 of (v[i] >> 7) << i; v[0] is the least-significant byte of "v".
// This is useful for testing/branching based on comparison results.
SIMD_INLINE uint32_t movemask(const scalar<uint8_t> v) { return v.raw >> 7; }

// Returns the most significant bit of each float/double lane (see above).
SIMD_INLINE uint32_t movemask(const scalar<float> v) {
  // Cannot return (v < 0) because +0.0 == -0.0.
  const Scalar<uint32_t> du;
  const auto bits = cast_to(du, v);
  return get_part(du, shift_right<31>(bits));
}
SIMD_INLINE uint32_t movemask(const scalar<double> v) {
  // Cannot return (v < 0) because +0.0 == -0.0.
  const Scalar<uint64_t> du;
  const auto bits = cast_to(du, v);
  return get_part(du, shift_right<63>(bits));
}

// Returns whether all lanes are equal to zero. Supported for all integer T.
template <typename T>
SIMD_INLINE bool all_zero(const scalar<T> v) {
  return v.raw == 0;
}

// Sum of all lanes, i.e. the only one.
template <typename T>
SIMD_INLINE scalar<T> sum_of_lanes(const scalar<T> v0) {
  return v0;
}

}  // namespace ext
}  // namespace pik

#endif  // PIK_SIMD_SCALAR_H_
